//*******************************************************
//
// KTR-10 Arduino 50MHz AM Transceiver Controler
//
//  Arduino Pro or Pro Mini / ATmega328P 5V 16MHz or 3.3V 8MHz
//
//  2023.10.21 KAWAKAMI Yukio
//*******************************************************
#include <Arduino_BuiltIn.h>
#include <Wire.h>
#include <MsTimer2.h>
#include <EEPROM.h>

#include "main.h"
#include "config.h"
#include "lcd.h"
#include "timer.h"
#include "Si5351.h"

// グローバル変数定義
volatile uint32_t TickCount = 0;
volatile uint16_t Timer;
volatile uint16_t MenuTimer;
volatile uint8_t Timer01;
volatile uint8_t PttCount;
volatile uint16_t SqlchTimer;

volatile int8_t EncCount = 0;

SWITCH Switch[4] = {
  { PIN_SW1, 0, 0 },
  { PIN_SW2, 0, 0 },
  { PIN_SW3, 0, 0 },
  { PIN_SW4, 0, 0 }
};

uint32_t Freq;     // 周波数
int32_t FreqStep;  // チューニングステップ

uint8_t Mode = MODE_RX;  // 動作モード MODE_RX, MODE_TX
uint8_t SqlchLevel = 0;
uint8_t AmpON;  // AMPENが正論理・負論理のどちらでも使えるように

CONFIG Config;

//*******************************************************
//  setuo()
//*******************************************************
void setup() {

  // debug用シリアルモニタ設定
  Serial.begin(115200);
  Serial.write("\nKTR-10 " VERSION "\n");
  Serial.write("Build:" __DATE__ " " __TIME__ "\n");

  Wire.begin();  // I2C 初期化

  // LCD初期化
  LCD_init();
  cursor(CURSOR_OFF);

  // Si5351初期化
  si5351_init();
  si5351_setupPLL(SI5351_PLL_A, PLL_MULT, 0, 1);  // PLL_A=800MHz固定

  // Config読み込み
  load_conf(&Config);
  if (Config.mark != CONFIG_MARK) {
    init_conf(&Config);  // Config初期化
    save_conf(&Config);
    Serial.write("Init Config\n");
  }

  // ポート設定
  pinMode(PIN_SW1, INPUT_PULLUP);
  pinMode(PIN_SW2, INPUT_PULLUP);
  pinMode(PIN_SW3, INPUT_PULLUP);
  pinMode(PIN_SW4, INPUT_PULLUP);
  pinMode(ENC_A, INPUT_PULLUP);
  pinMode(ENC_B, INPUT_PULLUP);

  pinMode(PIN_RXEN, OUTPUT);
  pinMode(PIN_TXEN, OUTPUT);  // 2024.5.6 追加
  pinMode(PIN_AMPEN, OUTPUT);
  AmpON = digitalRead(PIN_AMPEN) & HIGH;  // 起動時にOFFになっていることが前提
  set_amp(OFF);                           // AMP OFF
  pinMode(PIN_AMPEN, OUTPUT);
  digitalWrite(PIN_TXEN, LOW);  // TX OFF
  digitalWrite(PIN_RXEN, LOW);  // RX OFF

  // A/D基準電圧
  analogReference(DEFAULT);  // 内部基準電圧 5V

  // タイマ・ロータリーエンコーダ関係初期設定
  init_timer();

  Freq = Config.freq;  // 周波数初期値
  FreqStep = 1000;     // 1kHzステップ
  locate(0, 0);
  print_freq(Freq);
  set_rx_freq(Freq);

  digitalWrite(PIN_RXEN, HIGH);  // RX ON
  delay(200);                    // 200ms
  set_amp(ON);
  inkey();  // スイッチのフラグを読み捨て
}


//*******************************************************
//  loop()
//*******************************************************
void loop() {
  int volt;
  char msg[20];
  static int rssi = 0;

  // ATmega328 スリープ 割り込みがあれば起きる
  sleep();

  // 電源電圧読み込み
  volt = read_volt();
#if F_CUP >= 13000000
  if (volt < 70) {  // 7.0V以下
#else
  if (volt < 96) {  // 9.6V以下
#endif
    // 電源OFF処理
    set_amp(OFF);
    // 出力ポートを無効にする
    pinMode(PIN_RXEN, INPUT);
    pinMode(PIN_TXEN, INPUT);
    pinMode(PIN_AMPEN, INPUT);

    Config.freq = Freq;
    save_conf(&Config);  // Config 保存
    Serial.write("Power OFF\n");
    cls();
    print("Power OFF");

    while (true) {  // 無限ループ
      sleep();
    }
  }

  // PTT処理
  if (Config.ptt_volt_high > 0) {  // ptt_volt_high が設定されていなければ送信動作を行わない
    uint16_t ptt_volt;
    ptt_volt = read_ptt_volt();  // AD_PTT 電圧読み込み、mV単位で返す

    if ((ptt_volt <= Config.ptt_volt_high) && (ptt_volt >= Config.ptt_volt_low)) {
      if (PttCount == 0) PttCount = 1;
    } else {
      PttCount = 0;
    }

    if ((PttCount >= 50) && (Mode == MODE_RX)) {                            // 50ms以上連続して PTT ON
      if ((Freq >= (FREQ_BOTTOM + 3000)) && (Freq <= (FREQ_TOP - 3000))) {  // 上限・下限 3kHz付近は送信不可
        Mode = MODE_TX;
        set_amp(OFF);
        digitalWrite(PIN_RXEN, LOW);  // RX OFF
        set_rx_freq(0);
        delay(100);                    // 100ms
        digitalWrite(PIN_TXEN, HIGH);  // TX ON
        delay(100);                    // 100ms
        set_tx_freq(Freq);
      }
    } else if ((PttCount == 0) && (Mode == MODE_TX)) {
      Mode = MODE_RX;
      digitalWrite(PIN_TXEN, LOW);  // TX OFF
      set_tx_freq(0);
      delay(100);                    // 100ms
      digitalWrite(PIN_RXEN, HIGH);  // RX ON
      set_rx_freq(Freq);
      delay(200);  // 200ms
      set_amp(ON);
      inkey();  // スイッチのフラグを読み捨て
    }
  }

  if (Mode == MODE_TX) return;  // 以下、送信時はスイッチやエンコーダは動作しない

  if (!Timer01) {  // 0.1秒ごとの処理
    static int count = 0;
    static uint16_t last_rssi_ad = 0;
    uint16_t rssi_ad;

    Timer01 = 100;  // 100ms

    // RSSIの A/D値を簡単な LPF処理する
    rssi_ad = analogRead(AD_RSSI);
#if F_CPU < 13000000
    rssi_ad *= 33;
    rssi_ad /= 50;  // 基準電圧が 3.3Vなので 5V換算
#endif
    last_rssi_ad *= 3;
    last_rssi_ad /= 4;
    last_rssi_ad += (rssi_ad + 2) / 4;
    rssi = last_rssi_ad;

    // NJM2552 RSSI計算 変数rssiは static宣言されている
    rssi -= 51;          // 0.25V 時の A/D値
    rssi *= (70 - 10);   // 70〜 10 dBμV
    rssi /= (410 - 51);  // 2.0〜0.25V 時の A/D値
    rssi += 10;          // 10dBμV

    if (SqlchLevel > 0) {
      if (rssi >= SqlchLevel) {
        SqlchTimer = (uint16_t)Config.sqlch_delay * 1000;
        set_amp(ON);
      } else if (SqlchTimer == 0) {
        set_amp(OFF);
      }
    }

    if (++count >= 5) {  // 0.5秒ごとにレベルメーター表示
      count = 0;
      print_level_meter(rssi);  // レベルメーター、RS、RSSI表示
    }
  }

  if (read_SW(SW1)) {
    menu();  // メニュー
    cls();
    print_freq(Freq);
    set_rx_freq(Freq);
  }
  if (read_SW(SW2)) {
    // 未使用
  }
  if (read_SW(SW3)) {
    // オートスケルチ
    if (SqlchLevel == 0) {
      SqlchLevel = rssi + Config.sqlch_level;  // RSSIが現在のレベル＋sqlch_level 以上になったらスケルチが開く
      set_amp(OFF);
    } else {
      set_amp(ON);
      SqlchLevel = 0;  // スケルチ動作停止
    }
  }
  if (read_SW(SW4)) {
    menu_tune_step();  // チューニングステップ変更
    cls();
    print_freq(Freq);
  }

  // ロータリーエンコーダ処理
  int enc = read_ENC();
  if (enc != 0) {            // エンコーダが変化した
    Freq += enc * FreqStep;  // エンコーダ変化分読み込み
    if (Freq < FREQ_BOTTOM) Freq = FREQ_BOTTOM;
    if (Freq > FREQ_TOP) Freq = FREQ_TOP;

    // FreqStepの値に切りそろえる
    if (enc < 0) Freq += FreqStep - 1;
    Freq /= FreqStep;
    Freq *= FreqStep;

    set_rx_freq(Freq);
    locate(0, 0);
    print_freq(Freq);
  }

  if (!Timer) {  // 1秒ごとの処理
    Timer = 1000;

    locate(11, 0);
    sprintf(msg, "%2d.", volt / 10);
    print(msg);
    volt %= 10;
    sprintf(msg, "%dV", volt);
    print(msg);

    // debug 動作モニタ
    //locate(11, 1);
    //sprintf(msg, "%5d", TickCount / 1000);
    //print(msg);
  }
}


//*******************************************************
//  オーディオアンプ ON/OFF
//  アンプのICの Enable/Disableが正論理・負論理のどちらでも使えるように
//  ただし、Disableの状態で起動するようにプルダウン・プルアップされていることが前提
//*******************************************************
void set_amp(int flag) {

  if (flag) {
    digitalWrite(PIN_AMPEN, AmpON);  // AMP ON
  } else {
    digitalWrite(PIN_AMPEN, AmpON ^ HIGH);  // AMP OFF
  }
}


//*******************************************************
//  電源電圧読み込み
//  戻り値は 0.1V単位
//*******************************************************
int read_volt(void) {

  int32_t v = (int32_t)analogRead(AD_VOLT);

#if F_CPU >= 13000000
  v *= 50;          // 5.0V
  v *= (100 + 47);  // 抵抗分圧 R5 + R6
  v /= 47;          // R6
#else
  v *= 33;          // 3.3V
  v *= (100 + 22);  // 抵抗分圧 R1 + R2
  v /= 22;          // R2
#endif

  v += 512;
  v /= 1024;
  return (int)v;  // 0.1V単位  max 15.6V
}


//*******************************************************
//  PTT 電圧読み込み
//  戻り値は mV単位
//*******************************************************
int read_ptt_volt(void) {
  uint16_t ptt_volt;

  ptt_volt = analogRead(AD_PTT);

#if F_CPU >= 13000000
  ptt_volt *= 49;  // mVの電圧値変換
  ptt_volt /= 10;  // ptt_volt * 5000 / 1023
#else
  ptt_volt *= 32;   // mVの電圧値変換
  ptt_volt /= 10;   // ptt_volt * 3300 / 1023
#endif

  return (int)ptt_volt;
}


//*******************************************************
//  レベルメーターと RSSI値表示
//*******************************************************
void print_level_meter(int rssi) {
  int rs;
  int s;
  char c;
  char msg[20];

  if (rssi > 99) rssi = 99;
  if (rssi < Config.rs9_level) {
    rs = 9 - ((int)(Config.rs9_level + 3) - rssi) / 3;
  } else {
    rs = (rssi - Config.rs9_level) / 4 + 9;  // AGCが効くので
  }
  if (rs < 0) rs = 0;
  if (rs > 12) rs = 12;

  locate(0, 1);

  s = 0;
  for (int i = 0; i < 4; i++) {
    if (rs <= s) {
      lcd_putc(' ');
    } else if (rs >= (s + 3)) {
      lcd_putc(2);  // 3本
    } else {
      lcd_putc(rs - s - 1);
    }
    s += 3;
  }

  c = ' ';
  if (rs > 9) {
    rs = 9;
    c = '+';
  }
  sprintf(msg, "S%d%c %2ddB", rs, c, rssi);
  locate(4, 1);
  print(msg);
}

//*******************************************************
//	周波数文字列作成
//*******************************************************
void set_freq_str(char msg[], uint32_t freq) {
  char buff[8];

  if (freq == 0) {
    strcpy(msg, "--------");
  } else if (freq < (100L * 1000 * 1000)) {  // 100MHz以下
    sprintf(buff, "%6ld", freq / 100);
    msg[0] = buff[0];
    msg[1] = buff[1];
    msg[2] = '.';
    msg[3] = buff[2];
    msg[4] = buff[3];
    msg[5] = buff[4];
    msg[6] = '.';
    msg[7] = buff[5];
    msg[8] = 0;
  } else {  // 100MHz以上
    sprintf(buff, "%6ld", freq / 1000);
    msg[0] = buff[0];
    msg[1] = buff[1];
    msg[2] = buff[2];
    msg[3] = '.';
    msg[4] = buff[3];
    msg[5] = 'M';
    msg[6] = 'H';
    msg[7] = 'z';
    msg[8] = 0;
  }
}


//*******************************************************
//	カーソル位置に周波数表示
//*******************************************************
void print_freq(uint32_t freq) {
  char msg[10];

  set_freq_str(msg, freq);
  print(msg);
}


//*******************************************************
//  Si5351 PLL出力設定
//  freq = 0 で出力OFF
//*******************************************************
void set_freq(int ch, uint32_t freq) {

  if (freq == 0) {
    si5351_set_out(ch, 0);
  } else {
    si5351_set_frequency_fixedpll(ch, SI5351_PLL_A, PLL_FREQ, freq);
    si5351_set_out(ch, 1);
  }
}


//*******************************************************
//  受信周波数設定
//  freq = 0 で出力OFF
//*******************************************************
void set_rx_freq(uint32_t freq) {
  int32_t f;

  if (freq == 0) {
    set_freq(0, 0);
    return;
  }

  f = freq - Config.if_freq + (int32_t)Config.pll_offset + Config.rx_offset;
  set_freq(0, f);

  locate(13, 1);
  print(" RX");
}


//*******************************************************
//  送信周波数設定
//  freq = 0 で出力OFF
//*******************************************************
void set_tx_freq(uint32_t freq) {

  if (freq == 0) {
    set_freq(1, 0);
    return;
  }

  set_freq(1, freq + (int32_t)Config.pll_offset + Config.tx_offset);

  locate(13, 1);
  print(" TX");
}


//*******************************************************
//  menuからの呼び出し用
//  pll_offset調整用コールバック
//*******************************************************
void pll_offset_callback(int offset) {

  if (Mode == MODE_RX) {
    set_freq(0, Freq - Config.if_freq + (int32_t)offset * 10 + Config.rx_offset);
  } else if (Mode == MODE_TX) {
    set_freq(1, Freq + (int32_t)offset * 10 + Config.tx_offset);
  }
}


//*******************************************************
//  menuからの呼び出し用
//  rx_offset調整用コールバック
//*******************************************************
void rx_offset_callback(int offset) {

  if (Mode == MODE_RX) {
    set_freq(0, Freq - Config.if_freq + (int32_t)Config.pll_offset + offset * 10);
  }
}

//*******************************************************
//  menuからの呼び出し用
//  tx_offset調整用コールバック
//*******************************************************
void tx_offset_callback(int offset) {
  if (Mode == MODE_TX) {
    set_freq(1, Freq + (int32_t)Config.pll_offset + offset * 10);
  }
}